package io.flutter.plugins.nfc_manager

import android.annotation.TargetApi
import android.nfc.NdefMessage
import android.nfc.NdefRecord
import android.nfc.NfcAdapter
import android.nfc.Tag
import android.nfc.tech.IsoDep
import android.nfc.tech.MifareClassic
import android.nfc.tech.MifareUltralight
import android.nfc.tech.Ndef
import android.nfc.tech.NdefFormatable
import android.nfc.tech.NfcA
import android.nfc.tech.NfcB
import android.nfc.tech.NfcBarcode
import android.nfc.tech.NfcF
import android.nfc.tech.NfcV
import android.nfc.tech.TagTechnology
import android.os.Build
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import io.flutter.plugin.common.PluginRegistry.Registrar
import java.io.IOException
import java.util.*

class NfcManagerPlugin(private val mRegistrar: Registrar, private val mChannel: MethodChannel): MethodCallHandler {
    private val mAdapter = NfcAdapter.getDefaultAdapter(mRegistrar.context())
    private val mCachedTags = mutableMapOf<String, Tag>()
    private var mConnectedTechnology: TagTechnology? = null

    companion object {
        @JvmStatic
        fun registerWith(registrar: Registrar) {
            val channel = MethodChannel(registrar.messenger(), "plugins.flutter.io/nfc_manager")
            channel.setMethodCallHandler(NfcManagerPlugin(registrar, channel))
        }
    }

    override fun onMethodCall(call: MethodCall, result: Result) {
        when (call.method) {
            "startSession" -> handleStartSession(result)
            "stopSession" -> handleStopSession(result)
            "readNdef" -> handleReadNdef(result, call.argument("key")!!)
            "writeNdef" -> handleWriteNdef(result, call.argument("key")!!, call.argument("message")!!)
            "writeLock" -> handleWriteLock(result, call.argument("key")!!)
            "dispose" -> handleDispose(result, call.argument("key")!!)
            else -> result.notImplemented()
        }
    }

    private fun handleStartSession(result: Result) {
        if (Build.VERSION.SDK_INT < Build.VERSION_CODES.KITKAT) {
            result.error("", "Requires KITKAT", null)
        } else {
            mAdapter.enableReaderMode(mRegistrar.activity(), { tag ->
                val key = UUID.randomUUID().toString()
                mCachedTags[key] = tag
                mRegistrar.activity().runOnUiThread { mChannel.invokeMethod("onTagDiscovered", utilMapFromTag(tag, key)) }
            }, utilGetFlag(), null)
            result.success(true)
        }
    }

    private fun handleStopSession(result: Result) {
        if (Build.VERSION.SDK_INT < Build.VERSION_CODES.KITKAT) {
            result.error("", "Requires KITKAT", null)
        } else {
            mAdapter.disableReaderMode(mRegistrar.activity())
            result.success(true)
        }
    }

    private fun handleReadNdef(result: Result, key: String) {
        val tag = mCachedTags[key] ?: run {
            result.error("", "Tag is not found", null)
            return
        }

        val ndef = Ndef.get(tag) ?: run {
            result.error("", "Tag is not ndef", null)
            return
        }

        try {
            forceConnect(ndef)
            ndef.ndefMessage?.let {
                result.success(utilMapFromNdefMessage(it))
            } ?: run {
                result.success(null)
            }
        } catch (e: IOException) {
            result.error("", e.localizedMessage, null)
        }
    }

    private fun handleWriteNdef(result: Result, key: String, message: Map<String, Any?>) {
        val tag = mCachedTags[key] ?: run {
            result.error("", "Tag is not found", null)
            return
        }

        val ndef = Ndef.get(tag) ?: run {
            result.error("", "Tag is not ndef", null)
            return
        }

        try {
            forceConnect(ndef)
            ndef.writeNdefMessage(utilNdefMessageFromMap(message))
            result.success(true)
        } catch (e: IOException) {
            result.error("", e.localizedMessage, null)
        }
    }

    private fun handleWriteLock(result: Result, key: String) {
        val tag = mCachedTags[key] ?: run {
            result.error("", "Tag is not found", null)
            return
        }

        val ndef = Ndef.get(tag) ?: run {
            result.error("", "Tag is not ndef", null)
            return
        }

        try {
            forceConnect(ndef)
            ndef.makeReadOnly()
            result.success(true)
        } catch (e: IOException) {
            result.error("", e.localizedMessage, null)
        }
    }

    private fun handleDispose(result: Result, key: String) {
        mCachedTags.remove(key)?.let { tag ->
            mConnectedTechnology?.let { technology ->
                if (technology.tag == tag && technology.isConnected) {
                    try { technology.close() } catch (e: IOException) { /* Do nothing */ }
                    mConnectedTechnology = null
                }
            }
        }

        result.success(true)
    }

    @Throws(IOException::class)
    private fun forceConnect(technology: TagTechnology) {
        mConnectedTechnology?.let { connectedTechnology ->
            if (connectedTechnology.tag == technology.tag && connectedTechnology::class.java.name == connectedTechnology::class.java.name) return
            try { connectedTechnology.close() } catch (e: IOException) { /* Do nothing */ }
            technology.connect()
            mConnectedTechnology = technology
        } ?: run {
            technology.connect()
            mConnectedTechnology = technology
        }
    }

    @TargetApi(Build.VERSION_CODES.KITKAT)
    private fun utilGetFlag(): Int {
        return NfcAdapter.FLAG_READER_NFC_A or NfcAdapter.FLAG_READER_NFC_B or NfcAdapter.FLAG_READER_NFC_F or NfcAdapter.FLAG_READER_NFC_V
    }

    private fun utilMapFromTag(tag: Tag, key: String): Map<String, Any?> {
        val map = mutableMapOf<String, Any?>(
                "key" to key,
                "id" to tag.id
        )
        tag.techList.map { tech ->
            val techKey = tech.split(".").last().toLowerCase(Locale.ROOT) // android.nfc.tech.NfcA -> nfca
            val technology = utilTechnologyFromTag(tag, tech)
            val techValue = if (technology == null) mapOf() else utilMapFromTechnology(technology)
            map[techKey] = techValue
        }
        return map
    }

    private fun utilNdefMessageFromMap(map: Map<String, Any?>): NdefMessage {
        val records = (map["records"] as List<*>).filterIsInstance<Map<String, Any?>>()
        return NdefMessage(records.map { utilNdefRecordFromMap(it) }.toTypedArray())
    }

    private fun utilNdefRecordFromMap(map: Map<String, Any?>): NdefRecord {
        return NdefRecord(
                (map["typeNameFormat"] as Int).toShort(),
                map["type"] as ByteArray,
                map["id"] as? ByteArray,
                map["payload"] as ByteArray
        )
    }

    private fun utilMapFromNdefMessage(message: NdefMessage): Map<String, Any?> {
        return mapOf(
                "byteArrayLength" to message.byteArrayLength,
                "records" to message.records.map { utilMapFromNdefRecord(it) }.toList()
        )
    }

    private fun utilMapFromNdefRecord(record: NdefRecord): Map<String, Any?> {
        return mapOf(
                "id" to record.id,
                "payload" to record.payload,
                "type" to record.type,
                "typeNameFormat" to record.tnf
        )
    }

    private fun utilTechnologyFromTag(tag: Tag, tech: String): TagTechnology? {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.JELLY_BEAN_MR1 && tech == NfcBarcode::class.java.name) {
            return NfcBarcode.get(tag)
        }

        return when (tech) {
            NfcA::class.java.name -> NfcA.get(tag)
            NfcB::class.java.name -> NfcB.get(tag)
            NfcF::class.java.name -> NfcF.get(tag)
            NfcV::class.java.name -> NfcV.get(tag)
            IsoDep::class.java.name -> IsoDep.get(tag)
            MifareClassic::class.java.name -> MifareClassic.get(tag)
            MifareUltralight::class.java.name -> MifareUltralight.get(tag)
            Ndef::class.java.name -> Ndef.get(tag)
            NdefFormatable::class.java.name -> NdefFormatable.get(tag)
            else -> null
        }
    }

    private fun utilMapFromTechnology(technology: TagTechnology): Map<String, Any?> {
        return when (technology) {
            is NfcA -> mapOf()
            is NfcB -> mapOf()
            is NfcF -> mapOf()
            is NfcV -> mapOf()
            is IsoDep -> mapOf()
            is MifareClassic -> mapOf()
            is MifareUltralight -> mapOf()
            is Ndef -> mapOf(
                    "cachedNdefMessage" to if (technology.cachedNdefMessage == null) null else utilMapFromNdefMessage(technology.cachedNdefMessage),
                    "canMakeReadOnly" to technology.canMakeReadOnly(),
                    "isWritable" to technology.isWritable,
                    "maxSize" to technology.maxSize,
                    "type" to technology.type
            )
            is NdefFormatable -> mapOf()
            else -> mapOf()
        }
    }
}
