import 'dart:io';

import 'package:flutter/services.dart';

import './nfc_ndef.dart';

const _channel = MethodChannel('plugins.flutter.io/nfc_manager');

class NfcManager {
  NfcManager._() {
    _channel.setMethodCallHandler(_handleMethodCall);
  }

  static NfcManager _instance;
  static NfcManager get instance => _instance ??= NfcManager._();

  void Function(NfcTag) _onTagDiscovered;

  Future<bool> startSession({
    void Function(NfcTag) onTagDiscovered,
  }) {
    _onTagDiscovered = onTagDiscovered;
    return _channel.invokeMethod('startSession');
  }

  Future<bool> stopSession() {
    _onTagDiscovered = null;
    return _channel.invokeMethod('stopSession');
  }

  Future<void> _handleMethodCall(MethodCall call) async {
    switch (call.method) {
      case 'onTagDiscovered':
        _handleOnTagDiscovered(call.arguments);
        break;
      default:
        throw('Not implemented: ${call.method}');
    }
  }

  void _handleOnTagDiscovered(dynamic arguments) {
    final tag = _createNfcTag(Map<String, dynamic>.from(arguments));
    if (_onTagDiscovered != null)
      _onTagDiscovered(tag);
    tag._dispose();
  }

  NfcTag _createNfcTag(Map<String, dynamic> data) {
    String key = data.remove('key');

    NfcNdef ndef;
    if (data.containsKey('ndef')) {
      final ndefData = Map<String, dynamic>.from(data.remove('ndef'));
      final cachedNdefMessageData = Map<String, dynamic>.from(ndefData.remove('cachedNdefMessage'));
      final cachedNdefMessage = _nfcNdefMessageFromJson(cachedNdefMessageData);
      ndef = NfcNdef._(key, cachedNdefMessage, ndefData);
    }

    return NfcTag._(key, ndef, data);
  }
}

class NfcTag {
  NfcTag._(this._key, this.ndef, this.additionalData);

  final String _key;

  final NfcNdef ndef;

  final Map<String, dynamic> additionalData;

  Future<bool> _dispose() {
    return _channel.invokeMethod('dispose', {'key': _key});
  }
}

class NfcNdef {
  NfcNdef._(this._tagKey, this.cachedNdefMessage, this.additionalData);

  final String _tagKey;

  final NfcNdefMessage cachedNdefMessage;

  final Map<String, dynamic> additionalData;

  Future<bool> writeNdef(NfcNdefMessage message) {
    assert(Platform.isAndroid);
    return _channel.invokeMethod('writeNdef', {
      'key': _tagKey,
      'message': _nfcNdefMessageToJson(message),
    });
  }

  Future<bool> writeLock() {
    assert(Platform.isAndroid);
    return _channel.invokeMethod('writeLock', {'key': _tagKey});
  }
}

NfcNdefMessage _nfcNdefMessageFromJson(Map<String, dynamic> json) {
  return NfcNdefMessage(
    (json['records'] as List).map((e) =>
      _nfcNdefRecordFromJson(Map<String, dynamic>.from(e))
    ).toList());
}

NfcNdefRecord _nfcNdefRecordFromJson(Map<String, dynamic> json) {
  return NfcNdefRecord(
    json['id'],
    json['payload'],
    json['type'],
    json['typeNameFormat'],
  );
}

Map<String, dynamic> _nfcNdefMessageToJson(NfcNdefMessage instance) {
  return {'records': instance.records.map(_nfcNdefRecordToJson).toList()};
}

Map<String, dynamic> _nfcNdefRecordToJson(NfcNdefRecord instance) {
  return {
    'id': instance.id,
    'payload': instance.payload,
    'type': instance.type,
    'typeNameFormat': instance.typeNameFormat,
  };
}
